@@ -1,7 +1,7 @@
 /*******************************************************************************
    This file is part of Distributed Privacy Guard (DKGPG).
 
- Copyright (C) 2017  Heiko Stamer <HeikoStamer@gmx.net>
+ Copyright (C) 2017, 2018  Heiko Stamer <HeikoStamer@gmx.net>
 
    DKGPG is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -68,9 +68,12 @@ std::vector<size_t>			dkg_qual;
 std::vector<mpz_ptr>			dkg_v_i;
 std::vector< std::vector<mpz_ptr> >	dkg_c_ik;
 gcry_mpi_t 				dsa_p, dsa_q, dsa_g, dsa_y, dsa_x, elg_p, elg_q, elg_g, elg_y, elg_x;
+gcry_mpi_t				dsa_r, dsa_s, elg_r, elg_s, rsa_n, rsa_e, rsa_md;
 gcry_mpi_t 				gk, myk, sig_r, sig_s;
+gcry_mpi_t				revdsa_r, revdsa_s, revelg_r, revelg_s, revrsa_md;
 
 int 					opt_verbose = 0;
+bool					libgcrypt_secmem = false;
 bool					opt_binary = false;
 char					*opt_ifilename = NULL;
 char					*opt_ofilename = NULL;
@@ -144,19 +147,26 @@ void compute_decryption_share
 	// Advances in Cryptology - EUROCRYPT '97, LNCS 1233, pp. 103--118, 1997.
 
 	// compute the decryption share
-	mpz_t nizk_gk, r_i, R;
-	mpz_init(nizk_gk), mpz_init(r_i), mpz_init(R);
+	mpz_t nizk_gk, r_i, R, foo;
+	mpz_init(nizk_gk), mpz_init(r_i), mpz_init(R), mpz_init(foo);
 	mpz_spowm(R, dkg->g, dkg->x_i, dkg->p);
 	if (mpz_cmp(R, dkg->v_i[dkg->i]))
 	{
 		std::cerr << "ERROR: check of DKG public verification key failed" << std::endl;
-		mpz_clear(nizk_gk), mpz_clear(r_i), mpz_clear(R);
+		mpz_clear(nizk_gk), mpz_clear(r_i), mpz_clear(R), mpz_clear(foo);
 		exit(-1);
 	}
 	if (!mpz_set_gcry_mpi(gk, nizk_gk))
 	{
 		std::cerr << "ERROR: converting message component failed" << std::endl;
-		mpz_clear(nizk_gk), mpz_clear(r_i), mpz_clear(R);
+		mpz_clear(nizk_gk), mpz_clear(r_i), mpz_clear(R), mpz_clear(foo);
+		exit(-1);
+	}
+	mpz_powm(foo, nizk_gk, dkg->q, dkg->p); // additional check for subgroup property
+	if (mpz_cmp_ui(foo, 1L))
+	{
+		std::cerr << "ERROR: (g^k)^q equiv 1 mod p not satisfied" << std::endl;
+		mpz_clear(nizk_gk), mpz_clear(r_i), mpz_clear(R), mpz_clear(foo);
 		exit(-1);
 	}
 	mpz_spowm(r_i, nizk_gk, dkg->x_i, dkg->p);
@@ -183,19 +193,26 @@ void compute_decryption_share
 	std::ostringstream dds;
 	dds << "dds|" << dkg->i << "|" << r_i << "|" << c << "|" << r << "|";
 	mpz_clear(c), mpz_clear(r), mpz_clear(c2), mpz_clear(a), mpz_clear(b), mpz_clear(omega);
-	mpz_clear(nizk_gk), mpz_clear(r_i), mpz_clear(R);
+	mpz_clear(nizk_gk), mpz_clear(r_i), mpz_clear(R), mpz_clear(foo);
 	result = dds.str();
 }
 
 void prove_decryption_share_interactive_publiccoin
 	(GennaroJareckiKrawczykRabinDKG *dkg, mpz_srcptr r_i, aiounicast *aiou, CachinKursawePetzoldShoupRBC *rbc, JareckiLysyanskayaEDCF *edcf, std::ostream &err)
 {
-	mpz_t nizk_gk;
-	mpz_init(nizk_gk);
+	mpz_t nizk_gk, foo;
+	mpz_init(nizk_gk), mpz_init(foo);
 	if (!mpz_set_gcry_mpi(gk, nizk_gk))
 	{
 		std::cerr << "ERROR: converting message component failed" << std::endl;
-		mpz_clear(nizk_gk);
+		mpz_clear(nizk_gk), mpz_clear(foo);
+		exit(-1);
+	}
+	mpz_powm(foo, nizk_gk, dkg->q, dkg->p); // additional check for subgroup property
+	if (mpz_cmp_ui(foo, 1L))
+	{
+		std::cerr << "ERROR: (g^k)^q equiv 1 mod p not satisfied" << std::endl;
+		mpz_clear(nizk_gk), mpz_clear(foo);
 		exit(-1);
 	}
 	// set ID for RBC
@@ -224,7 +241,7 @@ void prove_decryption_share_interactive_publiccoin
 	}
 	// release
 	mpz_clear(c), mpz_clear(r), mpz_clear(c2), mpz_clear(a), mpz_clear(b), mpz_clear(omega);
-	mpz_clear(nizk_gk);
+	mpz_clear(nizk_gk), mpz_clear(foo);
 	// unset ID for RBC
 	rbc->unsetID();
 }
@@ -1173,6 +1190,24 @@ int main
 			std::cout << peers[i] << std::endl;
 	}
 
+	// lock memory
+	if (!lock_memory())
+	{
+		std::cerr << "WARNING: locking memory failed; CAP_IPC_LOCK required for memory protection" << std::endl;
+		// at least try to use libgcrypt's secure memory
+		if (!gcry_check_version(TMCG_LIBGCRYPT_VERSION))
+		{
+			std::cerr << "ERROR: libgcrypt version >= " << TMCG_LIBGCRYPT_VERSION << " required" << std::endl;
+			return -1;
+		}
+		gcry_control(GCRYCTL_SUSPEND_SECMEM_WARN);
+		gcry_control(GCRYCTL_USE_SECURE_RNDPOOL);
+		gcry_control(GCRYCTL_INIT_SECMEM, 16384, 0);
+		gcry_control(GCRYCTL_RESUME_SECMEM_WARN);
+		gcry_control(GCRYCTL_INITIALIZATION_FINISHED, 0);
+		libgcrypt_secmem = true;
+	}
+
 	// initialize LibTMCG
 	if (!init_libTMCG())
 	{
@@ -1260,7 +1295,7 @@ int main
 			dds_input.push_back(dds[i]);
 		std::string dds_radix;
 		CallasDonnerhackeFinneyShawThayerRFC4880::Radix64Encode(dds_input, dds_radix, false);
-		std::cout << "My decryption share (keep confidential): " << dds_radix << std::endl;
+		std::cout << "Your decryption share (keep confidential): " << dds_radix << std::endl;
 		mpz_init(r_i), mpz_init(c), mpz_init(r);
 		if (!verify_decryption_share(dkg, dds, idx, r_i, c, r))
 		{
