@@ -1,7 +1,7 @@
 /*******************************************************************************
    This file is part of Distributed Privacy Guard (DKGPG).
 
- Copyright (C) 2017  Heiko Stamer <HeikoStamer@gmx.net>
+ Copyright (C) 2017, 2018  Heiko Stamer <HeikoStamer@gmx.net>
 
    DKGPG is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -66,9 +66,12 @@ std::vector<size_t>			dkg_qual;
 std::vector<mpz_ptr>			dkg_v_i;
 std::vector< std::vector<mpz_ptr> >	dkg_c_ik;
 gcry_mpi_t 				dsa_p, dsa_q, dsa_g, dsa_y, dsa_x, elg_p, elg_q, elg_g, elg_y, elg_x;
+gcry_mpi_t				dsa_r, dsa_s, elg_r, elg_s, rsa_n, rsa_e, rsa_md;
 gcry_mpi_t 				gk, myk, sig_r, sig_s;
+gcry_mpi_t				revdsa_r, revdsa_s, revelg_r, revelg_s, revrsa_md;
 
 int 					opt_verbose = 0;
+bool					libgcrypt_secmem = false;
 char					*opt_crs = NULL;
 char					*opt_passwords = NULL;
 char					*opt_hostname = NULL;
@@ -398,7 +401,10 @@ void run_instance
 		delete dkg, delete dss, delete rbc, delete vtmf, delete aiou, delete aiou2;
 		exit(-1);
 	}
-	x = gcry_mpi_new(2048);
+	if (libgcrypt_secmem)
+		x = gcry_mpi_snew(2048);
+	else	
+		x = gcry_mpi_new(2048);
 	if (!mpz_get_gcry_mpi(x, dsa_x))
 	{
 		std::cerr << "P_" << whoami << ": mpz_get_gcry_mpi() failed for dsa_x" << std::endl;
@@ -505,7 +511,10 @@ void run_instance
 				c_ik[j].push_back(tmp);
 			}
 		}
-		x_i = gcry_mpi_new(2048);
+		if (libgcrypt_secmem)
+			x_i = gcry_mpi_snew(2048);
+		else		
+			x_i = gcry_mpi_new(2048);
 		if (!mpz_get_gcry_mpi(x_i, dss->x_i))
 		{
 			std::cerr << "P_" << whoami << ": mpz_get_gcry_mpi() failed for dss->x_i" << std::endl;
@@ -533,7 +542,10 @@ void run_instance
 			delete dkg, delete dss, delete rbc, delete vtmf, delete aiou, delete aiou2;
 			exit(-1);
 		}
-		xprime_i = gcry_mpi_new(2048);
+		if (libgcrypt_secmem)
+			xprime_i = gcry_mpi_snew(2048);
+		else
+			xprime_i = gcry_mpi_new(2048);
 		if (!mpz_get_gcry_mpi(xprime_i, dss->xprime_i))
 		{
 			std::cerr << "P_" << whoami << ": mpz_get_gcry_mpi() failed for dss->xprime_i" << std::endl;
@@ -823,7 +835,10 @@ void run_instance
 				c_ik[j].push_back(tmp);
 			}
 		}
-		x_i = gcry_mpi_new(2048);
+		if (libgcrypt_secmem)
+			x_i = gcry_mpi_snew(2048);
+		else
+			x_i = gcry_mpi_new(2048);
 		if (!mpz_get_gcry_mpi(x_i, dkg->x_i))
 		{
 			std::cerr << "P_" << whoami << ": mpz_get_gcry_mpi() failed for dkg->x_i" << std::endl;
@@ -849,7 +864,10 @@ void run_instance
 			delete dkg, delete dss, delete rbc, delete vtmf, delete aiou, delete aiou2;
 			exit(-1);
 		}
-		xprime_i = gcry_mpi_new(2048);
+		if (libgcrypt_secmem)
+			xprime_i = gcry_mpi_snew(2048);
+		else
+			xprime_i = gcry_mpi_new(2048);
 		if (!mpz_get_gcry_mpi(xprime_i, dkg->xprime_i))
 		{
 			std::cerr << "P_" << whoami << ": mpz_get_gcry_mpi() failed for dkg->xprime_i" << std::endl;
@@ -1026,7 +1044,7 @@ void run_instance
 	all.insert(all.end(), uidsig.begin(), uidsig.end());
 	all.insert(all.end(), sub.begin(), sub.end());
 	all.insert(all.end(), subsig.begin(), subsig.end());
-	CallasDonnerhackeFinneyShawThayerRFC4880::ArmorEncode(6, all, armor);
+	CallasDonnerhackeFinneyShawThayerRFC4880::ArmorEncode(TMCG_OPENPGP_ARMOR_PUBLIC_KEY_BLOCK, all, armor);
 	if (opt_verbose > 1)
 		std::cout << armor << std::endl;
 	std::ofstream pubofs((pubfilename.str()).c_str(), std::ofstream::out);
@@ -1054,7 +1072,7 @@ void run_instance
 	all.insert(all.end(), uidsig.begin(), uidsig.end());
 	all.insert(all.end(), ssb.begin(), ssb.end());
 	all.insert(all.end(), subsig.begin(), subsig.end());
-	CallasDonnerhackeFinneyShawThayerRFC4880::ArmorEncode(5, all, armor);
+	CallasDonnerhackeFinneyShawThayerRFC4880::ArmorEncode(TMCG_OPENPGP_ARMOR_PRIVATE_KEY_BLOCK, all, armor);
 	if (opt_verbose > 1)
 		std::cout << armor << std::endl;
 	std::ofstream secofs((secfilename.str()).c_str(), std::ofstream::out);
@@ -1647,6 +1665,24 @@ int main
 			std::cout << peers[i] << std::endl;
 	}
 
+	// lock memory
+	if (!lock_memory())
+	{
+		std::cerr << "WARNING: locking memory failed; CAP_IPC_LOCK required for memory protection" << std::endl;
+		// at least try to use libgcrypt's secure memory
+		if (!gcry_check_version(TMCG_LIBGCRYPT_VERSION))
+		{
+			std::cerr << "ERROR: libgcrypt version >= " << TMCG_LIBGCRYPT_VERSION << " required" << std::endl;
+			return -1;
+		}
+		gcry_control(GCRYCTL_SUSPEND_SECMEM_WARN);
+		gcry_control(GCRYCTL_USE_SECURE_RNDPOOL);
+		gcry_control(GCRYCTL_INIT_SECMEM, 16384, 0);
+		gcry_control(GCRYCTL_RESUME_SECMEM_WARN);
+		gcry_control(GCRYCTL_INITIALIZATION_FINISHED, 0);
+		libgcrypt_secmem = true;
+	}
+
 	// initialize LibTMCG
 	if (!init_libTMCG())
 	{
@@ -1946,6 +1982,7 @@ int main
 #endif
 
 	std::cout << "INFO: running local test with " << peers.size() << " participants" << std::endl;
+	std::cerr << "WARNING: due to cache issues the generated shares are identical, don't use them!" << std::endl;
 	// open pipes
 	for (size_t i = 0; i < peers.size(); i++)
 	{
