@@ -1,7 +1,7 @@
 /*******************************************************************************
    This file is part of Distributed Privacy Guard (DKGPG).
 
- Copyright (C) 2017  Heiko Stamer <HeikoStamer@gmx.net>
+ Copyright (C) 2017, 2018  Heiko Stamer <HeikoStamer@gmx.net>
 
    DKGPG is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -68,13 +68,19 @@ std::vector<size_t>			dkg_qual;
 std::vector<mpz_ptr>			dkg_v_i;
 std::vector< std::vector<mpz_ptr> >	dkg_c_ik;
 gcry_mpi_t 				dsa_p, dsa_q, dsa_g, dsa_y, dsa_x, elg_p, elg_q, elg_g, elg_y, elg_x;
+gcry_mpi_t				dsa_r, dsa_s, elg_r, elg_s, rsa_n, rsa_e, rsa_md;
 gcry_mpi_t 				gk, myk, sig_r, sig_s;
+gcry_mpi_t				revdsa_r, revdsa_s, revelg_r, revelg_s, revrsa_md;
 
 int 					opt_verbose = 0;
+bool					libgcrypt_secmem = false;
 char					*opt_passwords = NULL;
 char					*opt_hostname = NULL;
 unsigned long int			opt_p = 55000, opt_W = 5;
 
+mpz_t					cache[TMCG_MAX_SSRANDOMM_CACHE], cache_mod;
+size_t					cache_avail = 0;
+
 void run_instance
 	(size_t whoami, const size_t num_xtests)
 {
@@ -229,7 +235,7 @@ void run_instance
 	std::stringstream err_log;
 	if (opt_verbose)
 		std::cout << "R_" << whoami << ": dss.Refresh()" << std::endl;
-	if (!dss->Refresh(peers.size(), whoami, idx2dkg, dkg2idx, aiou, rbc, err_log, false))
+	if (!dss->Refresh(peers.size(), whoami, idx2dkg, dkg2idx, aiou, rbc, err_log, false, cache, cache_mod, &cache_avail))
 	{
 		std::cerr << "R_" << whoami << ": " << "tDSS Refresh() failed" << std::endl;
 		std::cerr << "R_" << whoami << ": log follows " << std::endl << err_log.str();
@@ -303,7 +309,10 @@ void run_instance
 		release_mpis();
 		exit(-1);
 	}
-	x_i = gcry_mpi_new(2048);
+	if (libgcrypt_secmem)
+		x_i = gcry_mpi_snew(2048);
+	else
+		x_i = gcry_mpi_new(2048);
 	if (!mpz_get_gcry_mpi(x_i, dss->x_i))
 	{
 		std::cerr << "R_" << whoami << ": mpz_get_gcry_mpi() failed for x_i" << std::endl;
@@ -317,7 +326,10 @@ void run_instance
 		release_mpis();
 		exit(-1);
 	}
-	xprime_i = gcry_mpi_new(2048);
+	if (libgcrypt_secmem)
+		xprime_i = gcry_mpi_snew(2048);
+	else
+		xprime_i = gcry_mpi_new(2048);
 	if (!mpz_get_gcry_mpi(xprime_i, dss->xprime_i))
 	{
 		std::cerr << "R_" << whoami << ": mpz_get_gcry_mpi() failed for xprime_i" << std::endl;
@@ -417,7 +429,7 @@ void run_instance
 	all.insert(all.end(), uidsig.begin(), uidsig.end());
 	all.insert(all.end(), ssb.begin(), ssb.end());
 	all.insert(all.end(), subsig.begin(), subsig.end());
-	CallasDonnerhackeFinneyShawThayerRFC4880::ArmorEncode(5, all, armor);
+	CallasDonnerhackeFinneyShawThayerRFC4880::ArmorEncode(TMCG_OPENPGP_ARMOR_PRIVATE_KEY_BLOCK, all, armor);
 	if (opt_verbose > 1)
 		std::cout << armor << std::endl;
 	std::ofstream secofs((secfilename.str()).c_str(), std::ofstream::out | std::ofstream::trunc);
@@ -686,6 +698,24 @@ int main
 			std::cout << peers[i] << std::endl;
 	}
 
+	// lock memory
+	if (!lock_memory())
+	{
+		std::cerr << "WARNING: locking memory failed; CAP_IPC_LOCK required for memory protection" << std::endl;
+		// at least try to use libgcrypt's secure memory
+		if (!gcry_check_version(TMCG_LIBGCRYPT_VERSION))
+		{
+			std::cerr << "ERROR: libgcrypt version >= " << TMCG_LIBGCRYPT_VERSION << " required" << std::endl;
+			return -1;
+		}
+		gcry_control(GCRYCTL_SUSPEND_SECMEM_WARN);
+		gcry_control(GCRYCTL_USE_SECURE_RNDPOOL);
+		gcry_control(GCRYCTL_INIT_SECMEM, 16384, 0);
+		gcry_control(GCRYCTL_RESUME_SECMEM_WARN);
+		gcry_control(GCRYCTL_INITIALIZATION_FINISHED, 0);
+		libgcrypt_secmem = true;
+	}
+
 	// initialize LibTMCG
 	if (!init_libTMCG())
 	{
@@ -695,6 +725,55 @@ int main
 	if (opt_verbose)
 		std::cout << "INFO: using LibTMCG version " << version_libTMCG() << std::endl;
 
+	// initialize cache
+	std::string armored_pubkey = "undefined";
+	if (opt_hostname != NULL)
+	{
+		if (!read_key_file(hostname + "_dkg-pub.asc", armored_pubkey))
+			armored_pubkey = "undefined";
+	}
+	else
+	{
+		for (size_t i = 0; i < peers.size(); i++)
+		{
+			if (read_key_file(peers[i] + "_dkg-pub.asc", armored_pubkey))
+				break;
+		}
+	}
+	if (armored_pubkey == "undefined")
+	{
+		std::cerr << "ERROR: no corresponding public key file found" << std::endl;
+		return -1;
+	}
+	init_mpis();
+	time_t ckeytime = 0, ekeytime = 0, csubkeytime = 0, esubkeytime = 0;
+	tmcg_byte_t keyusage = 0, keystrength = 1;
+	if (!parse_public_key(armored_pubkey, ckeytime, ekeytime, csubkeytime, esubkeytime, keyusage, keystrength))
+	{
+		std::cerr << "ERROR: cannot parse the provided public key" << std::endl;
+		release_mpis();
+		return -1;
+	}
+	if (!keystrength)
+	{
+		std::cerr << "ERROR: provided public key is too weak" << std::endl;
+		release_mpis();
+		return -1;
+	}
+	if (!mpz_set_gcry_mpi(dsa_q, dss_q))
+	{
+		std::cerr << "ERROR: mpz_set_gcry_mpi() failed for dss_q" << std::endl;
+		release_mpis();
+		return -1;
+	}
+	std::cout << "We need some entropy to cache very strong randomness for share refresh." << std::endl;
+	std::cout << "Please use other programs, move the mouse, and type on your keyboard: " << std::endl; 
+	mpz_ssrandomm_cache_init(cache, cache_mod, &cache_avail, (2 * peers.size()), dss_q);
+	std::cout << "Thank you!" << std::endl;
+	keyid.clear(), subkeyid.clear(), pub.clear(), sub.clear(), uidsig.clear(), subsig.clear();
+	dss_qual.clear(), dss_x_rvss_qual.clear(), dss_c_ik.clear(), dkg_qual.clear(), dkg_v_i.clear(), dkg_c_ik.clear();
+	release_mpis();
+
 	// initialize return code
 	int ret = 0;
 	// create underlying point-to-point channels, if built-in TCP/IP service requested
@@ -714,6 +793,8 @@ int main
 		ret = tcpip_io();
 		tcpip_close();
 		tcpip_done();
+		// release cache
+		mpz_ssrandomm_cache_done(cache, cache_mod, &cache_avail);
 		return ret;
 	}
 
@@ -765,6 +846,8 @@ int main
 	};
 	ret = GNUNET_PROGRAM_run(argc, argv, usage, about, myoptions, &gnunet_run, argv[0]);
 	GNUNET_free((void *) argv);
+	// release cache
+	mpz_ssrandomm_cache_done(cache, cache_mod, &cache_avail);
 	if (ret == GNUNET_OK)
 		return 0;
 	else
@@ -774,6 +857,7 @@ int main
 #endif
 
 	std::cout << "INFO: running local test with " << peers.size() << " participants" << std::endl;
+	std::cerr << "WARNING: due to cache issues the generated shares are identical, don't use them!" << std::endl;
 	// open pipes
 	for (size_t i = 0; i < peers.size(); i++)
 	{
@@ -826,6 +910,9 @@ int main
 		}
 	}
 	
+	// release cache
+	mpz_ssrandomm_cache_done(cache, cache_mod, &cache_avail);
+	// finish	
 	return ret;
 }
 
