@@ -1,7 +1,7 @@
 /*******************************************************************************
    This file is part of Distributed Privacy Guard (DKGPG).
 
- Copyright (C) 2017  Heiko Stamer <HeikoStamer@gmx.net>
+ Copyright (C) 2017, 2018  Heiko Stamer <HeikoStamer@gmx.net>
 
    DKGPG is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -27,6 +27,7 @@
 #include <map>
 #include <string>
 #include <algorithm>
+#include <ctime>
 
 #include <libTMCG.hh>
 
@@ -47,9 +48,12 @@ std::vector<size_t>			dkg_qual;
 std::vector<mpz_ptr>			dkg_v_i;
 std::vector< std::vector<mpz_ptr> >	dkg_c_ik;
 gcry_mpi_t 				dsa_p, dsa_q, dsa_g, dsa_y, dsa_x, elg_p, elg_q, elg_g, elg_y, elg_x;
+gcry_mpi_t				dsa_r, dsa_s, elg_r, elg_s, rsa_n, rsa_e, rsa_md;
 gcry_mpi_t 				gk, myk, sig_r, sig_s;
+gcry_mpi_t				revdsa_r, revdsa_s, revelg_r, revelg_s, revrsa_md;
 
 int 					opt_verbose = 0;
+bool					libgcrypt_secmem = false;
 bool 					opt_binary = false;
 char					*opt_ifilename = NULL;
 
@@ -59,6 +63,10 @@ int main
 	static const char *usage = "dkg-verify [OPTIONS] -i INPUTFILE KEYFILE";
 	static const char *about = PACKAGE_STRING " " PACKAGE_URL;
 	static const char *version = PACKAGE_VERSION " (" PACKAGE_NAME ")";
+	char *opt_sigfrom = NULL, *opt_sigto = NULL;
+	std::string sigfrom_str, sigto_str;
+	struct tm sigfrom_tm = { 0 }, sigto_tm = { 0 };
+	time_t sigfrom = 1243810800, sigto = time(NULL);
 
 	// parse command line arguments
 	for (size_t i = 0; i < (size_t)(argc - 1); i++)
@@ -74,6 +82,26 @@ int main
 			}
 			continue;
 		}
+		else if (arg.find("-f") == 0)
+		{
+			size_t idx = ++i;
+			if ((arg.find("-f") == 0) && (idx < (size_t)(argc - 1)) && (opt_sigfrom == NULL))
+			{
+				sigfrom_str = argv[i+1];
+				opt_sigfrom = (char*)sigfrom_str.c_str();
+			}
+			continue;
+		}
+		else if (arg.find("-t") == 0)
+		{
+			size_t idx = ++i;
+			if ((arg.find("-t") == 0) && (idx < (size_t)(argc - 1)) && (opt_sigto == NULL))
+			{
+				sigto_str = argv[i+1];
+				opt_sigto = (char*)sigto_str.c_str();
+			}
+			continue;
+		}
 		else if ((arg.find("--") == 0) || (arg.find("-b") == 0) || (arg.find("-v") == 0) || (arg.find("-h") == 0) || (arg.find("-V") == 0))
 		{
 			if ((arg.find("-h") == 0) || (arg.find("--help") == 0))
@@ -82,8 +110,10 @@ int main
 				std::cout << about << std::endl;
 				std::cout << "Arguments mandatory for long options are also mandatory for short options." << std::endl;
 				std::cout << "  -b, --binary   consider KEYFILE as binary input" << std::endl;
+				std::cout << "  -f TIMESPEC    signature made before given TIMESPEC is not valid" << std::endl;
 				std::cout << "  -h, --help     print this help" << std::endl;
 				std::cout << "  -i FILENAME    verify detached signature on FILENAME" << std::endl;
+				std::cout << "  -t TIMESPEC    signature made after given TIMESPEC is not valid" << std::endl;
 				std::cout << "  -v, --version  print the version number" << std::endl;
 				std::cout << "  -V, --verbose  turn on verbose output" << std::endl;
 				return 0; // not continue
@@ -118,6 +148,28 @@ int main
 		std::cerr << "ERROR: some filename missing; usage: " << usage << std::endl;
 		return -1;
 	}
+	if (opt_sigfrom)
+	{
+		strptime(opt_sigfrom, "%Y-%m-%d_%H:%M:%S", &sigfrom_tm);
+		sigfrom = mktime(&sigfrom_tm);
+		if (sigfrom == ((time_t) -1))
+		{
+			perror("dkg-verify (mktime)");
+			std::cerr << "ERROR: cannot convert TIMESPEC; required format: YYYY-MM-DD_HH:MM:SS" << std::endl;
+			return -1;
+		}
+	}
+	if (opt_sigto)
+	{
+		strptime(opt_sigto, "%Y-%m-%d_%H:%M:%S", &sigto_tm);
+		sigto = mktime(&sigto_tm);
+		if (sigto == ((time_t) -1))
+		{
+			perror("dkg-verify (mktime)");
+			std::cerr << "ERROR: cannot convert TIMESPEC; required format: YYYY-MM-DD_HH:MM:SS" << std::endl;
+			return -1;
+		}
+	}
 
 	// initialize LibTMCG
 	if (!init_libTMCG())
@@ -130,15 +182,22 @@ int main
 
 	// read and parse the public key (no ElGamal subkey required)
 	std::string armored_pubkey;
-	if (opt_binary && !read_binary_key_file(kfilename, 6, armored_pubkey))
+	if (opt_binary && !read_binary_key_file(kfilename, TMCG_OPENPGP_ARMOR_PUBLIC_KEY_BLOCK, armored_pubkey))
 		return -1;
 	if (!opt_binary && !read_key_file(kfilename, armored_pubkey))
 		return -1;
 	init_mpis();
-	time_t ckeytime = 0, ekeytime = 0;
-	if (!parse_public_key(armored_pubkey, ckeytime, ekeytime, false))
+	time_t ckeytime = 0, ekeytime = 0, csubkeytime = 0, esubkeytime = 0;
+	tmcg_byte_t keyusage = 0, keystrength = 1;
+	if (!parse_public_key(armored_pubkey, ckeytime, ekeytime, csubkeytime, esubkeytime, keyusage, keystrength, false))
 	{
-		std::cerr << "ERROR: cannot parse the provided public key" << std::endl;
+		std::cerr << "ERROR: cannot parse resp. use the provided public key" << std::endl;
+		release_mpis();
+		return -1;
+	}
+	if (!keystrength)
+	{
+		std::cerr << "ERROR: provided public key is too weak" << std::endl;
 		release_mpis();
 		return -1;
 	}
@@ -161,15 +220,64 @@ int main
 
 	// parse the signature
 	tmcg_octets_t trailer;
-	tmcg_byte_t hashalgo = 0;
+	tmcg_byte_t hashalgo = 0, sigstrength = 1;
 	time_t csigtime = 0, sigexptime = 0;
 	bool sigV3 = false;
-	if (!parse_signature(signature, 0x00, csigtime, sigexptime, hashalgo, trailer, sigV3))
+	if (!parse_signature(signature, 0x00, csigtime, sigexptime, hashalgo, trailer, sigV3, sigstrength))
 	{
-		std::cerr << "ERROR: cannot parse the provided signature" << std::endl;
+		std::cerr << "ERROR: cannot parse resp. use the provided signature" << std::endl;
 		release_mpis();
 		return -1;
 	}
+	if (!sigstrength)
+	{
+		std::cerr << "ERROR: provided signature is too weak" << std::endl;
+		release_mpis();
+		return -1;
+	}
+
+	// additional validity checks on key and signature
+	time_t current_time = time(NULL);
+	// 1. key validity time (signatures made before key creation or after key expiry are not valid)
+	if (csigtime < ckeytime)
+	{
+		std::cout << "ERROR: signature was made before key creation" << std::endl;
+		return -2;
+	}
+	if (ekeytime && (csigtime > (ckeytime + ekeytime)))
+	{
+		std::cout << "ERROR: signature was made after key expiry" << std::endl;
+		return -2;
+	}
+	// 2. signature validity time (expired signatures are not valid)
+	if (sigexptime && (current_time > (csigtime + sigexptime)))
+	{
+		std::cout << "ERROR: signature is expired" << std::endl;
+		return -2;
+	}
+	// 3. key usage flags (signatures made by keys not with the "signing" capability are not valid)
+	if ((keyusage & 0x02) != 0x02)
+	{
+		std::cout << "ERROR: corresponding key was not intented for signing" << std::endl;
+		return -2;
+	}
+	// 4. key validity time (expired keys are not valid)
+	if (ekeytime && (current_time > (ckeytime + ekeytime)))
+	{
+		std::cout << "ERROR: corresponding key is expired" << std::endl;
+		return -2;
+	}
+	// 5. signature time (signatures made before the sigfrom or after the sigto timespec are not valid)
+	if (csigtime < sigfrom)
+	{
+		std::cout << "ERROR: signature was made before given TIMESPEC" << std::endl;
+		return -2;
+	}
+	if (csigtime > sigto)
+	{
+		std::cout << "ERROR: signature was made after given TIMESPEC" << std::endl;
+		return -2;
+	}
 
 	// compute the hash of the input file
 	if (opt_verbose)
